/* See LICENSE file for copyright and license details. */
//g++ ./icon-menu.c -o ./icon-menu -lX11 -lImlib2
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <dirent.h>
#include <limits.h>
#include <sys/stat.h>
#include <locale.h>
#include <stdarg.h>
#include "Imlib2.h"

#define FG(dc, col)  ((col)[(dc)->invert ? ColBG : ColFG])
#define BG(dc, col)  ((col)[(dc)->invert ? ColFG : ColBG])

Visual                         *button_vis;
Colormap                       button_cm;
int                            button_depth;

enum { ColBG, ColFG, ColBorder, ColLast };

typedef struct {
	int x, y, w, h;
	Bool invert;
	Display *dpy;
	GC gc;
	Pixmap canvas;
	struct {
		int ascent;
		int descent;
		int height;
		XFontSet set;
		XFontStruct *xfont;
	} font;
} DC;  /* draw context */

unsigned long getcolor(DC *dc, const char *colstr);
void drawrect(DC *dc, int x, int y, unsigned int w, unsigned int h, Bool fill, unsigned long color);
void drawtext(DC *dc, const char *text, unsigned long col[ColLast]);
void drawtextn(DC *dc, const char *text, size_t n, unsigned long col[ColLast]);
void initfont(DC *dc, const char *fontstr);
void freedc(DC *dc);
DC *initdc(void);
void mapdc(DC *dc, Window win, unsigned int w, unsigned int h);
void resizedc(DC *dc, unsigned int w, unsigned int h);
int textnw(DC *dc, const char *text, size_t len);
int textw(DC *dc, const char *text);
void eprintf(const char *fmt, ...);
void weprintf(const char *fmt, ...);

const char *progname;

#define MAX(a, b)   ((a) > (b) ? (a) : (b))
#define MIN(a, b)   ((a) < (b) ? (a) : (b))
#define DEFFONT     "fixed"
static Bool loadfont(DC *dc, const char *fontstr);
#define INRECT(x,y,rx,ry,rw,rh) ((x) >= (rx) && (x) < (rx)+(rw) && (y) >= (ry) && (y) < (ry)+(rh))
#define MIN(a,b)                ((a) < (b) ? (a) : (b))
#define MAX(a,b)                ((a) > (b) ? (a) : (b))

typedef struct Item Item;
struct Item {
	char *text;
	Item *next;          /* traverses all items */
	Item *left, *right;  /* traverses matching items */
};

static void appenditem(Item *item, Item **list, Item **last);
static void calcoffsets(void);
static void drawmenu(void);
static char *fstrstr(const char *s, const char *sub);
static void grabkeyboard(void);
static void insert(const char *s, ssize_t n);
static void keypress(XKeyEvent *ev);
static void match(void);
static size_t nextrune(int incr);
static void paste(void);
static void readstdin(void);
static void run(void);
static void setup(void);

static char text[BUFSIZ];
static int bh, mw, mh;
static int inputw = 0;
static int lines = 0;
static int promptw;
static size_t cursor = 0;
static const char *font = NULL;
static const char *prompt = NULL;
static const char *normbgcolor = "#cccccc";
static const char *normfgcolor = "#000000";
static const char *selbgcolor  = "#0066ff";
static const char *selfgcolor  = "#ffffff";
static unsigned long normcol[ColLast];
static unsigned long selcol[ColLast];
static Atom utf8;
static Bool topbar = True;
static DC *dc;
static Item *items = NULL;
static Item *matches, *sel;
static Item *prev, *curr, *next;
static Window root, win, win1, win2;
static int show_menu,only_menu;
char *c_kill;
static int (*fstrncmp)(const char *, const char *, size_t) = strncmp;

void
drawrect(DC *dc, int x, int y, unsigned int w, unsigned int h, Bool fill, unsigned long color) {
	XRectangle r = { dc->x + x, dc->y + y, w, h };

	if(!fill) {
		r.width -= 1;
		r.height -= 1;
	}
	XSetForeground(dc->dpy, dc->gc, color);
	(fill ? XFillRectangles : XDrawRectangles)(dc->dpy, dc->canvas, dc->gc, &r, 1);
}


void
drawtext(DC *dc, const char *text, unsigned long col[ColLast]) {
	char buf[256];
	size_t n, mn;

	/* shorten text if necessary */
	n = strlen(text);
	for(mn = MIN(n, sizeof buf); textnw(dc, text, mn) > dc->w - dc->font.height/2; mn--)
		if(mn == 0)
			return;
	memcpy(buf, text, mn);
	if(mn < n)
		for(n = MAX(mn-3, 0); n < mn; buf[n++] = '.');

	drawrect(dc, 0, 0, dc->w, dc->h, True, BG(dc, col));
	drawtextn(dc, buf, mn, col);
}

void
drawtextn(DC *dc, const char *text, size_t n, unsigned long col[ColLast]) {
	int x, y;

	x = dc->x + dc->font.height/2;
	y = dc->y + dc->font.ascent+1;

	XSetForeground(dc->dpy, dc->gc, FG(dc, col));
	if(dc->font.set)
		XmbDrawString(dc->dpy, dc->canvas, dc->font.set, dc->gc, x, y, text, n);
	else {
		XSetFont(dc->dpy, dc->gc, dc->font.xfont->fid);
		XDrawString(dc->dpy, dc->canvas, dc->gc, x, y, text, n);
	}
}

void
eprintf(const char *fmt, ...) {
	va_list ap;

	fprintf(stderr, "%s: ", progname);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}

void
freedc(DC *dc) {
	if(dc->font.set)
		XFreeFontSet(dc->dpy, dc->font.set);
	if(dc->font.xfont)
		XFreeFont(dc->dpy, dc->font.xfont);
	if(dc->canvas)
		XFreePixmap(dc->dpy, dc->canvas);
	XFreeGC(dc->dpy, dc->gc);
	XCloseDisplay(dc->dpy);
	free(dc);
}

unsigned long
getcolor(DC *dc, const char *colstr) {
	Colormap cmap = DefaultColormap(dc->dpy, DefaultScreen(dc->dpy));
	XColor color;

	if(!XAllocNamedColor(dc->dpy, cmap, colstr, &color, &color))
		eprintf("cannot allocate color '%s'\n", colstr);
	return color.pixel;
}

DC *
initdc(void) {
	DC *dc;

	if(!setlocale(LC_CTYPE, "") || !XSupportsLocale())
		weprintf("no locale support\n");
	if(!(dc = (DC*)malloc(sizeof *dc)))
		eprintf("cannot malloc %u bytes\n", sizeof *dc);
	if(!(dc->dpy = XOpenDisplay(NULL)))
		eprintf("cannot open display\n");

	dc->gc = XCreateGC(dc->dpy, DefaultRootWindow(dc->dpy), 0, NULL);
	XSetLineAttributes(dc->dpy, dc->gc, 1, LineSolid, CapButt, JoinMiter);
	dc->font.xfont = NULL;
	dc->font.set = NULL;
	dc->canvas = None;
	return dc;
}

void
initfont(DC *dc, const char *fontstr) {
	if(!loadfont(dc, fontstr ? fontstr : DEFFONT)) {
		if(fontstr != NULL)
			weprintf("cannot load font '%s'\n", fontstr);
		if(fontstr == NULL || !loadfont(dc, DEFFONT))
			eprintf("cannot load font '%s'\n", DEFFONT);
	}
	dc->font.height = dc->font.ascent + dc->font.descent;
}

Bool
loadfont(DC *dc, const char *fontstr) {
	char *def, **missing;
	int i, n;

	if(!*fontstr)
		return False;
	if((dc->font.set = XCreateFontSet(dc->dpy, fontstr, &missing, &n, &def))) {
		char **names;
		XFontStruct **xfonts;

		n = XFontsOfFontSet(dc->font.set, &xfonts, &names);
		for(i = dc->font.ascent = dc->font.descent = 0; i < n; i++) {
			dc->font.ascent = MAX(dc->font.ascent, xfonts[i]->ascent);
			dc->font.descent = MAX(dc->font.descent, xfonts[i]->descent);
		}
	}
	else if((dc->font.xfont = XLoadQueryFont(dc->dpy, fontstr))) {
		dc->font.ascent = dc->font.xfont->ascent;
		dc->font.descent = dc->font.xfont->descent;
	}
	if(missing)
		XFreeStringList(missing);
	return (dc->font.set || dc->font.xfont);
}

void
mapdc(DC *dc, Window win, unsigned int w, unsigned int h) {
	XCopyArea(dc->dpy, dc->canvas, win, dc->gc, 0, 0, w, h, 0, 0);
}

void
resizedc(DC *dc, unsigned int w, unsigned int h) {
	if(dc->canvas)
		XFreePixmap(dc->dpy, dc->canvas);
	dc->canvas = XCreatePixmap(dc->dpy, DefaultRootWindow(dc->dpy), w, h,
	                           DefaultDepth(dc->dpy, DefaultScreen(dc->dpy)));
	dc->x = dc->y = 0;
	dc->w = w;
	dc->h = h;
	dc->invert = False;
}

int
textnw(DC *dc, const char *text, size_t len) {
	if(dc->font.set) {
		XRectangle r;

		XmbTextExtents(dc->font.set, text, len, NULL, &r);
		return r.width;
	}
	return XTextWidth(dc->font.xfont, text, len);
}

int
textw(DC *dc, const char *text) {
	return textnw(dc, text, strlen(text)) + dc->font.height;
}

void
weprintf(const char *fmt, ...) {
	va_list ap;

	fprintf(stderr, "%s: warning: ", progname);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}

#define CACHE ".dmenu_cache"

static void die(const char *s);
static int qstrcmp(const void *a, const void *b);
static void scan(void);

static char **d_items = NULL;
size_t d_count;
static const char *home, *path;

int
dmenu_patch_main(void) {
	if(!(home = getenv("HOME")))
		die("no $HOME");
	if(!(path = getenv("PATH")))
		die("no $PATH");
	if(chdir(home) < 0)
		die("chdir failed");
	scan();
	return EXIT_SUCCESS;
}

void
die(const char *s) {
	fprintf(stderr, "dmenu_path: %s\n", s);
	exit(EXIT_FAILURE);
}

int
qstrcmp(const void *a, const void *b) {
	return strcmp(*(const char **)a, *(const char **)b);
}

void
scan(void) {
	char buf[PATH_MAX];
	char *dir, *p;
	size_t i;
	struct dirent *ent;
	DIR *dp;
	FILE *cache;

	d_count = 0;
	if(!(p = strdup(path)))
		die("strdup failed");
	for(dir = strtok(p, ":"); dir; dir = strtok(NULL, ":")) {
		if(!(dp = opendir(dir)))
			continue;
		while((ent = readdir(dp))) {
			snprintf(buf, sizeof buf, "%s/%s", dir, ent->d_name);
			if(ent->d_name[0] == '.' || access(buf, X_OK) < 0)
				continue;
			if(!(d_items = (char**)realloc(d_items, ++d_count * sizeof *d_items)))
				die("malloc failed");
			if(!(d_items[d_count-1] = strdup(ent->d_name)))
				die("strdup failed");
		}
		closedir(dp);
	}
	qsort(d_items, d_count, sizeof *d_items, qstrcmp);
	if(!(cache = fopen(CACHE, "w")))
		die("open failed");
	for(i = 0; i < d_count; i++) {
		if(i > 0 && !strcmp(d_items[i], d_items[i-1]))
			continue;
		fprintf(cache,  "%s\n", d_items[i]);
//		fprintf(stdout, "%s\n", d_items[i]);
	} 
	fclose(cache);
	free(p); 
    
}


int
main(int argc, char *argv[]) {
    if ( argc == 2 ) 
    {
        c_kill=argv[1];
        only_menu=0;
    }
    else
    {
        c_kill=argv[0];
        only_menu=1;
    }
    dmenu_patch_main();
	progname = "dmenu";
	dc = initdc();
	initfont(dc, font);
	readstdin();
	setup();
	run();

	return EXIT_FAILURE;  /* should not reach */
}

void
appenditem(Item *item, Item **list, Item **last) {
	if(!*last)
		*list = item;
	else
		(*last)->right = item;
	item->left = *last;
	item->right = NULL;
	*last = item;
}

void
calcoffsets(void) {
	unsigned int i, n;

	if(lines > 0)
		n = lines * bh;
	else
		n = mw - (promptw + inputw + textw(dc, "<") + textw(dc, ">"));

	for(i = 0, next = curr; next; next = next->right)
		if((i += (lines > 0) ? bh : MIN(textw(dc, next->text), n)) > n)
			break;
	for(i = 0, prev = curr; prev && prev->left; prev = prev->left)
		if((i += (lines > 0) ? bh : MIN(textw(dc, prev->left->text), n)) > n)
			break;
}

void
drawmenu(void) {
	int curpos;
	Item *item;

	dc->x = 0;
	dc->y = 0;
	dc->h = bh;
	drawrect(dc, 0, 0, mw, mh, True, BG(dc, normcol));

	if(prompt) {
		dc->w = promptw;
		drawtext(dc, prompt, selcol);
		dc->x = dc->w;
	}
	dc->w = (lines > 0 || !matches) ? mw - dc->x : inputw;
	drawtext(dc, text, normcol);
	if((curpos = textnw(dc, text, cursor) + dc->h/2 - 2) < dc->w)
		drawrect(dc, curpos, 2, 1, dc->h - 4, True, FG(dc, normcol));

	if(lines > 0) {
		dc->w = mw - dc->x;
		for(item = curr; item != next; item = item->right) {
			dc->y += dc->h;
			drawtext(dc, item->text, (item == sel) ? selcol : normcol);
		}
	}
	else if(matches) {
		dc->x += inputw;
		dc->w = textw(dc, "<");
		if(curr->left)
			drawtext(dc, "<", normcol);
		for(item = curr; item != next; item = item->right) {
			dc->x += dc->w;
			dc->w = MIN(textw(dc, item->text), mw - dc->x - textw(dc, ">"));
			drawtext(dc, item->text, (item == sel) ? selcol : normcol);
		}
		dc->w = textw(dc, ">");
		dc->x = mw - dc->w;
		if(next)
			drawtext(dc, ">", normcol);
	}
	mapdc(dc, win, mw, mh);
}

char *
fstrstr(const char *s, const char *sub) {
	size_t len;

	for(len = strlen(sub); *s; s++)
		if(!fstrncmp(s, sub, len))
			return (char *)s;
	return NULL;
}

void
grabkeyboard(void) {
	int i;

	for(i = 0; i < 1000; i++) {
		if(!XGrabKeyboard(dc->dpy, root, True, GrabModeAsync, GrabModeAsync, CurrentTime))
			return;
		usleep(1000);
	}
	eprintf("cannot grab keyboard\n");
}

void
insert(const char *s, ssize_t n) {
	if(strlen(text) + n > sizeof text - 1)
		return;
	memmove(text + cursor + n, text + cursor, sizeof text - cursor - MAX(n, 0));
	if(n > 0)
		memcpy(text + cursor, s, n);
	cursor += n;
	match();
}

void
keypress(XKeyEvent *ev) {
	char buf[32];
	size_t len;
	KeySym ksym;

	char cmd[1024];
	len = strlen(text);
	XLookupString(ev, buf, sizeof buf, &ksym, NULL);
	if(ev->state & ControlMask) {
		switch(tolower(ksym)) {
		default:
			return;
		case XK_a:
			ksym = XK_Home;
			break;
		case XK_b:
			ksym = XK_Left;
			break;
		case XK_c:
			ksym = XK_Escape;
			break;
		case XK_d:
			ksym = XK_Delete;
			break;
		case XK_e:
			ksym = XK_End;
			break;
		case XK_f:
			ksym = XK_Right;
			break;
		case XK_h:
			ksym = XK_BackSpace;
			break;
		case XK_i:
			ksym = XK_Tab;
			break;
		case XK_j:
			ksym = XK_Return;
			break;
		case XK_k:  /* delete right */
			text[cursor] = '\0';
			match();
			break;
		case XK_n:
			ksym = XK_Down;
			break;
		case XK_p:
			ksym = XK_Up;
			break;
		case XK_u:  /* delete left */
			insert(NULL, 0 - cursor);
			break;
		case XK_w:  /* delete word */
			while(cursor > 0 && text[nextrune(-1)] == ' ')
				insert(NULL, nextrune(-1) - cursor);
			while(cursor > 0 && text[nextrune(-1)] != ' ')
				insert(NULL, nextrune(-1) - cursor);
			break;
		case XK_y:  /* paste selection */
			XConvertSelection(dc->dpy, XA_PRIMARY, utf8, utf8, win, CurrentTime);
			return;
		}
	}
	switch(ksym) {
	default:
		if(!iscntrl(*buf))
			insert(buf, strlen(buf));
		break;
	case XK_Delete:
		if(cursor == len)
			return;
		cursor = nextrune(+1);
	case XK_BackSpace:
		if(cursor > 0)
			insert(NULL, nextrune(-1) - cursor);
		break;
	case XK_End:
		if(cursor < len) {
			cursor = len;
			break;
		}
		while(next) {
			sel = curr = next;
			calcoffsets();
		}
		while(sel && sel->right)
			sel = sel->right;
		break;
	case XK_Escape:
		if (only_menu==0)
		{
			XUnmapWindow(dc->dpy, win);
			XUngrabKeyboard(dc->dpy,CurrentTime);
			show_menu=0;
        }
        else
        {
			exit(0);
		}
        break;
	case XK_Home:
		if(sel == matches) {
			cursor = 0;
			break;
		}
		sel = curr = matches;
		calcoffsets();
		break;
	case XK_Left:
		if(cursor > 0 && (!sel || !sel->left || lines > 0)) {
			cursor = nextrune(-1);
			break;
		}
		else if(lines > 0)
			return;
	case XK_Up:
		if(sel && sel->left && (sel = sel->left)->right == curr) {
			curr = prev;
			calcoffsets();
		}
		break;
	case XK_Next:
		if(!next)
			return;
		sel = curr = next;
		calcoffsets();
		break;
	case XK_Prior:
		if(!prev)
			return;
		sel = curr = prev;
		calcoffsets();
		break;
	case XK_Return:
	case XK_KP_Enter:
		if (sel && !(ev->state & ShiftMask))
		{
			sprintf(cmd,"%s &",sel->text);
		}
		else
		{
			sprintf(cmd,"%s &",text);
	    }
	    system(cmd);
	    printf("%s\n",cmd);
		if (only_menu==0)
		{
			XUnmapWindow(dc->dpy, win);
			XUngrabKeyboard(dc->dpy,CurrentTime);
			show_menu=0;
        }
        else
        {
			exit(1);
		}
        break;
	case XK_Right:
		if(cursor < len) {
			cursor = nextrune(+1);
			break;
		}
		else if(lines > 0)
			return;
	case XK_Down:
		if(sel && sel->right && (sel = sel->right) == next) {
			curr = next;
			calcoffsets();
		}
		break;
	case XK_Tab:
		if(!sel)
			return;
		strncpy(text, sel->text, sizeof text);
		cursor = strlen(text);
		match();
		break;
	}
	drawmenu();
}

void
match(void) {
	size_t len;
	Item *item, *itemend, *lexact, *lprefix, *lsubstr, *exactend, *prefixend, *substrend;

	len = strlen(text);
	matches = lexact = lprefix = lsubstr = itemend = exactend = prefixend = substrend = NULL;
	for(item = items; item; item = item->next)
		if(!fstrncmp(text, item->text, len + 1))
			appenditem(item, &lexact, &exactend);
		else if(!fstrncmp(text, item->text, len))
			appenditem(item, &lprefix, &prefixend);
		else if(fstrstr(item->text, text))
			appenditem(item, &lsubstr, &substrend);

	if(lexact) {
		matches = lexact;
		itemend = exactend;
	}
	if(lprefix) {
		if(itemend) {
			itemend->right = lprefix;
			lprefix->left = itemend;
		}
		else
			matches = lprefix;
		itemend = prefixend;
	}
	if(lsubstr) {
		if(itemend) {
			itemend->right = lsubstr;
			lsubstr->left = itemend;
		}
		else
			matches = lsubstr;
	}
	curr = prev = next = sel = matches;
	calcoffsets();
}

size_t
nextrune(int incr) {
	size_t n, len;

	len = strlen(text);
	for(n = cursor + incr; n >= 0 && n < len && (text[n] & 0xc0) == 0x80; n += incr);
	return n;
}

void
paste(void) {
	char *p, *q;
	int di;
	unsigned long dl;
	Atom da;

	XGetWindowProperty(dc->dpy, win, utf8, 0, (sizeof text / 4) + 1, False,
	                   utf8, &da, &di, &dl, &dl, (unsigned char **)&p);
	insert(p, (q = strchr(p, '\n')) ? q-p : strlen(p));
	XFree(p);
	drawmenu();
}

void
readstdin(void) {
	char buf[sizeof text],*p;
	Item *item, **end;
	FILE *cache;

	if(!(cache = fopen(CACHE, "r")))
		die("open failed");

	printf("%d\n",d_count);
	for(end = &items; fgets(buf, sizeof buf, cache); *end = item, end = &item->next) {
		if((p = strchr(buf, '\n')))
			*p = '\0';
		if(!(item = (Item*)malloc(sizeof *item)))
			eprintf("cannot malloc %u bytes\n", sizeof *item);
		if(!(item->text = strdup(buf)))
			eprintf("cannot strdup %u bytes\n", strlen(buf)+1);
		item->next = item->left = item->right = NULL;
		inputw = MAX(inputw, textw(dc, item->text));
	}
    fclose(cache);
}

void
run(void) {
    int    text_w, text_h;
    Imlib_Font font;
    Imlib_Image                 buffer3;
    Imlib_Color_Range            range;
	if (only_menu==0)
	{
		button_vis = DefaultVisual(dc->dpy, DefaultScreen(dc->dpy));
		button_depth = DefaultDepth(dc->dpy, DefaultScreen(dc->dpy));
		button_cm = DefaultColormap(dc->dpy, DefaultScreen(dc->dpy));
		imlib_add_path_to_font_path("/usr/share/fonts/truetype/wqy/");
		imlib_add_path_to_font_path("/usr/share/fonts/wenquanyi/wqy-zenhei/");
		font = imlib_load_font("wqy-zenhei.ttc/12");

		range = imlib_create_color_range();
		imlib_context_set_color_range(range);
		imlib_context_set_color(255, 255, 255, 255);
		imlib_add_color_to_color_range(0);
		/* add an orange color, semi-transparent 10 units from the first */
		imlib_context_set_color(255, 200, 10, 255);
		imlib_add_color_to_color_range(10);
		/* add black, fully transparent at the end 20 units away */
		imlib_context_set_color(70, 0, 0, 255);
		imlib_add_color_to_color_range(20);
		imlib_context_set_display(dc->dpy);
		imlib_context_set_visual(button_vis);
		imlib_context_set_colormap(button_cm);
		imlib_context_set_dither(1);
		imlib_context_set_blend(0);
		imlib_context_set_color_modifier(NULL);
		imlib_context_set_font(font);
		imlib_context_set_color(255, 200, 10, 255);
		buffer3 = imlib_create_image(200, 34);    
	}
	XEvent ev;
	while(!XNextEvent(dc->dpy, &ev))
		switch(ev.type) {
		case Expose:
			if (only_menu==0)
			{
				imlib_context_set_drawable(win1); 
				imlib_context_set_image(buffer3);
				imlib_free_image();
				buffer3 = imlib_create_image(200, 30);    
				imlib_context_set_image(buffer3);
				imlib_image_fill_color_range_rectangle(0,0, 240, 48,-45.0);
					 /* free it */
						//         imlib_free_color_range();
				imlib_context_set_image(buffer3);
				imlib_context_set_drawable(win1); 
				imlib_render_image_on_drawable(0, 0); 
				imlib_context_set_font(font);
				imlib_get_text_size("Menu", &text_w, &text_h);
				imlib_context_set_color(0, 0, 0, 255);
				imlib_text_draw((64-text_w)/2,(24-text_h)/2,"Menu");
				/* set the buffer image as our current image */
				imlib_context_set_image(buffer3);
				imlib_render_image_on_drawable(0, 0); 
				imlib_context_set_image(buffer3);

				imlib_context_set_drawable(win2); 
				imlib_context_set_image(buffer3);
				imlib_free_image();
				buffer3 = imlib_create_image(200, 30);    
				imlib_context_set_image(buffer3);
				imlib_image_fill_color_range_rectangle(0,0, 240, 48,-45.0);
				imlib_context_set_image(buffer3);
				imlib_context_set_drawable(win2); 
				imlib_render_image_on_drawable(0, 0); 
				imlib_context_set_font(font);
				imlib_get_text_size("Logout", &text_w, &text_h);
				imlib_context_set_color(0, 0, 0, 255);
				imlib_text_draw((64-text_w)/2,(24-text_h)/2,"Logout");
				/* set the buffer image as our current image */
				imlib_context_set_image(buffer3);
				imlib_render_image_on_drawable(0, 0); 
				imlib_context_set_image(buffer3);
            }
            if ((show_menu==1) or (only_menu==1))
            {
                if(ev.xexpose.count == 0)
                    drawmenu();
            }
			break;
		case KeyPress:
            if ((show_menu==1) or (only_menu==1))
                keypress(&ev.xkey);
			break;
		case SelectionNotify:
            if ((show_menu==1) or (only_menu==1))
            {
                if(ev.xselection.property == utf8)
                    paste();
            }
			break;
        case ButtonPress:
            if (only_menu==0)
            {
				if (ev.xmotion.window == win1)
				{
					grabkeyboard();
					XMapRaised(dc->dpy, win);
					text[0] = '\0';
					match();
					show_menu=1;
				}
				else if (ev.xmotion.window == win2)
				{
					char cmd[1024];
					sprintf(cmd,"killall %s",c_kill);
					system(cmd);
				}
            }
            break;
        case EnterNotify:
            if (only_menu==0)
            {
				XResizeWindow(dc->dpy, win1, 64, 24);
				XResizeWindow(dc->dpy, win2, 64, 24);
				imlib_context_set_drawable(win1); 
				imlib_context_set_image(buffer3);
				imlib_free_image();
				buffer3 = imlib_create_image(200, 30);    
				imlib_context_set_image(buffer3);
				imlib_image_fill_color_range_rectangle(0,0, 240, 48,-45.0);
					 /* free it */
						//         imlib_free_color_range();
				imlib_context_set_image(buffer3);
				imlib_context_set_drawable(win1); 
				imlib_render_image_on_drawable(0, 0); 
				imlib_context_set_font(font);
				imlib_get_text_size("Menu", &text_w, &text_h);
				imlib_context_set_color(0, 0, 0, 255);
				imlib_text_draw((64-text_w)/2,(24-text_h)/2,"Menu");
				/* set the buffer image as our current image */
				imlib_context_set_image(buffer3);
				imlib_render_image_on_drawable(0, 0); 
				imlib_context_set_image(buffer3);

				imlib_context_set_drawable(win2); 
				imlib_context_set_image(buffer3);
				imlib_free_image();
				buffer3 = imlib_create_image(200, 30);    
				imlib_context_set_image(buffer3);
				imlib_image_fill_color_range_rectangle(0,0, 240, 48,-45.0);
				imlib_context_set_image(buffer3);
				imlib_context_set_drawable(win2); 
				imlib_render_image_on_drawable(0, 0); 
				imlib_context_set_font(font);
				imlib_get_text_size("Logout", &text_w, &text_h);
				imlib_context_set_color(0, 0, 0, 255);
				imlib_text_draw((64-text_w)/2,(24-text_h)/2,"Logout");
				/* set the buffer image as our current image */
				imlib_context_set_image(buffer3);
				imlib_render_image_on_drawable(0, 0); 
				imlib_context_set_image(buffer3);
            }
            break;
        case LeaveNotify:
            if (only_menu==0)
            {
				XResizeWindow(dc->dpy, win1, 64, 4);
				XResizeWindow(dc->dpy, win2, 64, 4);
            }
            break;
		case VisibilityNotify:
			if(ev.xvisibility.state != VisibilityUnobscured)
				XRaiseWindow(dc->dpy, win);
			break;
		}
}

void
setup(void) {
	int x, y, screen;
	XSetWindowAttributes wa;
#ifdef XINERAMA
	int n;
	XineramaScreenInfo *info;
#endif
    int    text_w, text_h;
    Imlib_Font font;
    Imlib_Image                 buffer3;
    Imlib_Color_Range            range;
	if (only_menu==0)
	{
		button_vis = DefaultVisual(dc->dpy, DefaultScreen(dc->dpy));
		button_depth = DefaultDepth(dc->dpy, DefaultScreen(dc->dpy));
		button_cm = DefaultColormap(dc->dpy, DefaultScreen(dc->dpy));
		imlib_add_path_to_font_path("/usr/share/fonts/truetype/wqy/");
		imlib_add_path_to_font_path("/usr/share/fonts/wenquanyi/wqy-zenhei/");
		font = imlib_load_font("wqy-zenhei.ttc/12");

		range = imlib_create_color_range();
		imlib_context_set_color_range(range);
		imlib_context_set_color(255, 255, 255, 255);
		imlib_add_color_to_color_range(0);
		/* add an orange color, semi-transparent 10 units from the first */
		imlib_context_set_color(255, 200, 10, 255);
		imlib_add_color_to_color_range(10);
		/* add black, fully transparent at the end 20 units away */
		imlib_context_set_color(70, 0, 0, 255);
		imlib_add_color_to_color_range(20);
		imlib_context_set_display(dc->dpy);
		imlib_context_set_visual(button_vis);
		imlib_context_set_colormap(button_cm);
		imlib_context_set_dither(1);
		imlib_context_set_blend(0);
		imlib_context_set_color_modifier(NULL);
		imlib_context_set_font(font);
		imlib_context_set_color(255, 200, 10, 255);
		buffer3 = imlib_create_image(200, 34);    
	}
	screen = DefaultScreen(dc->dpy);
	root = RootWindow(dc->dpy, screen);
	utf8 = XInternAtom(dc->dpy, "UTF8_STRING", False);

	normcol[ColBG] = getcolor(dc, normbgcolor);
	normcol[ColFG] = getcolor(dc, normfgcolor);
	selcol[ColBG] = getcolor(dc, selbgcolor);
	selcol[ColFG] = getcolor(dc, selfgcolor);

	/* menu geometry */
	bh = dc->font.height + 2;
	lines = MAX(lines, 0);
	mh = (lines + 1) * bh;
#ifdef XINERAMA
	if((info = XineramaQueryScreens(dc->dpy, &n))) {
		int i, di;
		unsigned int du;
		Window dw;

		XQueryPointer(dc->dpy, root, &dw, &dw, &x, &y, &di, &di, &du);
		for(i = 0; i < n; i++)
			if((monitor == info[i].screen_number)
			|| (monitor < 0 && INRECT(x, y, info[i].x_org, info[i].y_org, info[i].width, info[i].height)))
				break;
		x = info[i].x_org;
		y = info[i].y_org + (topbar ? 0 : info[i].height - mh);
		mw = info[i].width;
		XFree(info);
	}
	else
#endif
	{
		x = 0;
		y = topbar ? 0 : DisplayHeight(dc->dpy, screen) - mh;
		mw = DisplayWidth(dc->dpy, screen);
	}
	/* menu window */
	wa.override_redirect = True;
	wa.background_pixmap = ParentRelative;
	if (only_menu==0)
	{
		win1 = XCreateWindow(dc->dpy, root, mw/2-35,32, 64, 24, 0,
							DefaultDepth(dc->dpy, screen), CopyFromParent,
							DefaultVisual(dc->dpy, screen),
							CWOverrideRedirect | CWBackPixmap | CWEventMask, &wa);
		win2 = XCreateWindow(dc->dpy, root, mw/2+35,32, 64, 24, 0,
							DefaultDepth(dc->dpy, screen), CopyFromParent,
							DefaultVisual(dc->dpy, screen),
							CWOverrideRedirect | CWBackPixmap | CWEventMask, &wa);
		XSelectInput(dc->dpy, win1,  ButtonPressMask | ButtonReleaseMask | ButtonMotionMask | PointerMotionMask | ExposureMask | EnterWindowMask | LeaveWindowMask);
		XSelectInput(dc->dpy, win2,  ButtonPressMask | ButtonReleaseMask | ButtonMotionMask | PointerMotionMask | ExposureMask | EnterWindowMask | LeaveWindowMask);
	}
	wa.event_mask = ExposureMask | KeyPressMask | VisibilityChangeMask;
	win = XCreateWindow(dc->dpy, root, x, y, mw, mh, 0,
	                    DefaultDepth(dc->dpy, screen), CopyFromParent,
	                    DefaultVisual(dc->dpy, screen),
	                    CWOverrideRedirect | CWBackPixmap | CWEventMask, &wa);

	resizedc(dc, mw, mh);
	inputw = MIN(inputw, mw/3);
	promptw = prompt ? textw(dc, prompt) : 0;
	if (only_menu==0)
	{
		XMapRaised(dc->dpy, win1);
		XMapRaised(dc->dpy, win2);
		XRaiseWindow(dc->dpy,win1);
		XMapWindow(dc->dpy, win2);

		imlib_context_set_drawable(win1); 
		imlib_context_set_image(buffer3);
		imlib_free_image();
		buffer3 = imlib_create_image(200, 30);    
		imlib_context_set_image(buffer3);
		imlib_image_fill_color_range_rectangle(0,0, 240, 48,-45.0);
			 /* free it */
				//         imlib_free_color_range();
		imlib_context_set_image(buffer3);
		imlib_context_set_drawable(win1); 
		imlib_render_image_on_drawable(0, 0); 
		imlib_context_set_font(font);
		imlib_get_text_size("Menu", &text_w, &text_h);
		imlib_context_set_color(0, 0, 0, 255);
		imlib_text_draw((64-text_w)/2,(24-text_h)/2,"Menu");
		/* set the buffer image as our current image */
		imlib_context_set_image(buffer3);
		imlib_render_image_on_drawable(0, 0); 
		imlib_context_set_image(buffer3);

		imlib_context_set_drawable(win2); 
		imlib_context_set_image(buffer3);
		imlib_free_image();
		buffer3 = imlib_create_image(200, 30);    
		imlib_context_set_image(buffer3);
		imlib_image_fill_color_range_rectangle(0,0, 240, 48,-45.0);
		imlib_context_set_image(buffer3);
		imlib_context_set_drawable(win2); 
		imlib_render_image_on_drawable(0, 0); 
		imlib_context_set_font(font);
		imlib_get_text_size("Logout", &text_w, &text_h);
		imlib_context_set_color(0, 0, 0, 255);
		imlib_text_draw((64-text_w)/2,(24-text_h)/2,"Logout");
		/* set the buffer image as our current image */
		imlib_context_set_image(buffer3);
		imlib_render_image_on_drawable(0, 0); 
		imlib_context_set_image(buffer3);
		
		show_menu=0;
	}
	else
	{
		grabkeyboard();
		XMapRaised(dc->dpy, win);
		text[0] = '\0';
		match();
	}
}

